from django.apps import apps
from django.conf import settings
from django.db import DEFAULT_DB_ALIAS, models
import pgactivity
import pgactivity.models

from pglock import config, utils


class PGTableQueryCompiler(pgactivity.models.PGTableQueryCompiler):
    def get_pid_clause(self):
        if isinstance(self.query.pids, str):
            return f"AND pid = ANY({self.query.pids})"
        else:
            return super().get_pid_clause()

    def get_ctes(self):
        ctes = super().get_ctes()

        if self.query.relations:
            models = [
                apps.get_model(model) if isinstance(model, str) else model
                for model in self.query.relations
            ]
            rel_name_clause = (
                "AND pg_class.relname IN ("
                + ", ".join(f"'{model._meta.db_table}'" for model in models)
                + ")"
            )
        else:
            rel_name_clause = ""

        with self.connection.cursor() as cursor:  # pragma: no cover
            if utils.pg_maj_version(cursor) >= 14:
                # Waitstart is available in pg 14 and up
                wait_start_clause = "waitstart AS wait_start, NOW() - waitstart AS wait_duration"
            else:
                wait_start_clause = "NULL AS wait_start, NULL AS wait_duration"

        lock_cte = rf"""
            _pglock_lock_cte AS (
                SELECT
                    pid AS id,
                    pid AS activity_id,
                    TRIM(
                        BOTH '_' FROM UPPER(
                            REGEXP_REPLACE(
                                REPLACE(mode, 'Lock', ''),
                                '([A-Z])','_\1',
                                'g'
                            )
                        )
                    ) AS mode,
                    CASE pg_class.relkind
                        WHEN 'r' THEN 'TABLE'::text
                        WHEN 'i' THEN 'INDEX'::text
                        WHEN 'S' THEN 'SEQUENCE'::text
                        WHEN 't' THEN 'TOAST'::text
                        WHEN 'v' THEN 'VIEW'::text
                        WHEN 'm' THEN 'MATERIALIZED_VIEW'::text
                        WHEN 'c' THEN 'COMPOSITE_TYPE'::text
                        WHEN 'f' THEN 'FOREIGN_TABLE'::text
                        WHEN 'p' THEN 'PARTITIONED_TABLE'::text
                        WHEN 'I' THEN 'PARTITIONED_INDEX'::text
                        ELSE pg_class.relkind::text
                    END AS rel_kind,
                    UPPER(locktype) as type,
                    pg_class.relname as rel_name,
                    granted,
                    {wait_start_clause}
                FROM pg_locks
                JOIN pg_database ON pg_database.oid = pg_locks.database
                JOIN pg_class ON pg_class.oid = pg_locks.relation
                WHERE
                    pg_database.datname = '{settings.DATABASES[self.using]["NAME"]}'
                    {self.get_pid_clause()}
                    {rel_name_clause}
            )
        """

        blocked_lock_cte = """
            _pglock_blocked_lock_cte AS (
                SELECT
                    *,
                    UNNEST(pg_blocking_pids(activity_id)) AS blocking_activity_id
                FROM _pglock_lock_cte
            )
        """

        ctes.extend([lock_cte, blocked_lock_cte])
        return ctes


class PGTableQuery(pgactivity.models.PGTableQuery):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.relations = None

    def get_compiler(self, *args, **kwargs):
        compiler = super().get_compiler(*args, **kwargs)
        compiler.__class__ = PGTableQueryCompiler
        return compiler

    def __chain(self, *args, **kwargs):
        clone = super().__chain(*args, **kwargs)
        clone.relations = self.relations
        return clone


class PGTableQuerySet(pgactivity.models.PGTableQuerySet):
    def __init__(self, model=None, query=None, using=None, hints=None):
        if query is None:
            query = PGTableQuery(model)

        super().__init__(model, query, using, hints)


class PGLockQuerySet(PGTableQuerySet):
    """The Queryset for the `PGLock` model."""

    def on(self, *relations):
        """Set the relations to filter against.

        Currently model names or classes are accepted.
        """
        qs = self._clone()
        qs.query.relations = relations
        return qs

    def cancel_activity(self):
        """Cancel all PIDs in the ``activity`` field of the filtered queryset"""
        pids = list(self.values_list("activity_id", flat=True).distinct())
        return pgactivity.cancel(*pids, using=self.db)

    def terminate_activity(self):
        """Terminate all PIDs in the ``activity`` field of the filtered queryset"""
        pids = list(self.values_list("activity_id", flat=True).distinct())
        return pgactivity.terminate(*pids, using=self.db)

    def config(self, name, **overrides):
        """
        Use a config name from ``settings.PGLOCK_CONFIGS``
        to apply filters. Config overrides can be provided
        in the keyword arguments.

        Args:
            name (str): Name of the config. Must be a key from ``settings.PGLOCK_CONFIGS``.
            **overrides: Any overrides to apply to the final config dictionary.

        Returns:
            dict: The configuration
        """
        qset = self

        cfg = config.get(name, **overrides)

        qset = qset.using(cfg.get("database", DEFAULT_DB_ALIAS))
        qset = qset.pid(*cfg.get("pids", []))
        qset = qset.on(*cfg.get("on", []))

        for f in cfg.get("filters", []) or []:
            key, val = f.split("=", 1)
            qset = qset.filter(**{key: val})

        return qset


class BasePGLock(pgactivity.models.PGTable):
    type = models.CharField(max_length=64)
    activity = models.ForeignKey("pgactivity.PGActivity", on_delete=models.DO_NOTHING)
    mode = models.CharField(max_length=64)
    granted = models.BooleanField()
    wait_start = models.DateTimeField(null=True)
    wait_duration = models.DurationField(null=True)
    rel_kind = models.CharField(max_length=32)
    rel_name = models.CharField(max_length=256)

    objects = PGLockQuerySet.as_manager()

    class Meta:
        abstract = True


class PGLock(BasePGLock):
    """
    Wraps Postgres's ``pg_locks`` view.

    Attributes:
        type (models.CharField): The type of lock. One of
            RELATION, EXTEND, FROZENID, PAGE, TUPLE, TRANSACTIONID, VIRTUALXID,
            SPECTOKEN, OBJECT, USERLOCK, or ADVISORY.
        activity (models.ForeignKey[pgactivity.PGActivity]): The activity
            from ``pg_stats_activity`` this lock references.
        mode (models.CharField): The mode of lock. One of
            ACCESS_SHARE, ROW_SHARE, ROW_EXCLUSIVE, SHARE_UPDATE_EXCLUSIVE,
            SHARE, SHARE_ROW_EXCLUSIVE, EXCLUSIVE, ACCESS_EXCLUSIVE.
        granted (models.BooleanField): ``True`` if the lock has been granted,
            ``False`` if the lock is blocked by another.
        wait_start (models.DateTimeField): When the lock started waiting. Only
            available in Postgres 14 and up.
        wait_duration (models.DurationField): How long the lock has been blocked.
            Only available in Postgres 14 and up
        rel_kind (models.CharField): The kind of relation being locked. One of
            TABLE, INDEX, SEQUENCE, TOAST, VIEW, MATERIALIZED_VIEW,
            COMPOSITE_TYPE, FOREIGN_TABLE, PARTITIONED_TABLE, or
            PARTITIONED_INDEX.
        rel_name (models.CharField): The name of the relation. E.g. the table name
            when ``rel_kind=TABLE``.
    """

    class Meta:
        managed = False
        db_table = "_pglock_lock_cte"
        default_manager_name = "no_objects"


class BlockedPGLockQuerySet(PGLockQuerySet):
    """The Queryset for the `BlockedPGLock` model. Inherits `PGLockQuerySet`"""

    def cancel_blocking_activity(self):
        """Cancel all PIDs in the ``blocking_activity`` field of the filtered queryset"""
        pids = list(self.values_list("blocking_activity_id", flat=True).distinct())
        return pgactivity.cancel(*pids, using=self.db)

    def terminate_blocking_activity(self):
        """Terminate all PIDs in the ``blocking_activity`` field of the filtered queryset"""
        pids = list(self.values_list("blocking_activity_id", flat=True).distinct())
        return pgactivity.terminate(*pids, using=self.db)

    def pid(self, *pids):
        qs = self._clone()

        if pids:
            qs.query.pids = " || ".join(f"{pid} || pg_blocking_pids({pid})" for pid in pids)
        else:  # pragma: no cover
            qs.query.pids = pids

        return qs


class BlockedPGLock(BasePGLock):
    """Models a blocked lock.

    Uses Postgres's ``pg_blocking_pids`` function to unnest and
    denormalize any blocking activity for a lock, returning both
    the activity and blocking activity as a row.

    Attributes:
        blocking_activity (models.ForeignKey[pgactivity.PGActivity]): The
            activity that's blocking the lock.
    """

    blocking_activity = models.ForeignKey(
        "pgactivity.PGActivity", on_delete=models.DO_NOTHING, related_name="+"
    )

    objects = BlockedPGLockQuerySet.as_manager()

    class Meta:
        managed = False
        db_table = "_pglock_blocked_lock_cte"
        default_manager_name = "no_objects"



import contextlib
import datetime as dt
import functools
import hashlib
import inspect
import threading

from django.apps import apps
from django.db import connections, DEFAULT_DB_ALIAS, transaction
from django.db.utils import OperationalError
import pgactivity

from pglock import utils

if utils.psycopg_maj_version == 2:
    import psycopg2.extensions
elif utils.psycopg_maj_version == 3:
    import psycopg.pq
else:
    raise AssertionError


# Lock levels
ACCESS_SHARE = "ACCESS SHARE"
ROW_SHARE = "ROW SHARE"
ROW_EXCLUSIVE = "ROW EXCLUSIVE"
SHARE_UPDATE_EXCLUSIVE = "SHARE UPDATE EXCLUSIVE"
SHARE = "SHARE"
SHARE_ROW_EXCLUSIVE = "SHARE ROW EXCLUSIVE"
EXCLUSIVE = "EXCLUSIVE"
ACCESS_EXCLUSIVE = "ACCESS EXCLUSIVE"


_timeout = threading.local()
_unset = object()


class SideEffect:
    """The base class for side effects"""


class Return(SideEffect):
    """
    The side effect for returning the lock status when
    using `pglock.advisory` or `pglock.model`.
    """


class Raise(SideEffect):
    """
    The side effect for raising an error on lock acquisition
    failure when using `pglock.advisory` or `pglock.model`.
    """


class Skip(SideEffect):
    """
    The side effect for skipping wrapped code on lock acquisition error
    when using `pglock.advisory` as a decorator.
    """


def _cast_timeout(timeout):
    if timeout is not None and timeout is not _unset:
        if isinstance(timeout, (int, float)):
            timeout = dt.timedelta(seconds=timeout)

        if not isinstance(timeout, dt.timedelta):
            raise TypeError("Must supply int, float, or timedelta to pglock.timeout")

        if timeout < dt.timedelta(milliseconds=1):
            timeout = dt.timedelta()

    return timeout


def _is_transaction_errored(cursor):
    """
    True if the current transaction is in an errored state
    """
    if utils.psycopg_maj_version == 2:
        return (
            cursor.connection.get_transaction_status()
            == psycopg2.extensions.TRANSACTION_STATUS_INERROR
        )
    elif utils.psycopg_maj_version == 3:
        return cursor.connection.info.transaction_status == psycopg.pq.TransactionStatus.INERROR
    else:
        raise AssertionError


@contextlib.contextmanager
def lock_timeout(timeout=_unset, *, using=DEFAULT_DB_ALIAS, **timedelta_kwargs):
    """Set the lock timeout as a decorator or context manager.

    A value of ``None`` will set an infinite lock timeout.
    A value of less than a millisecond is not permitted.

    Nested invocations will successfully apply and rollback the timeout to
    the previous value.

    Args:
        timeout (Union[datetime.timedelta, int, float, None]): The number
            of seconds as an integer or float. Use a timedelta object to
            precisely specify the timeout interval. Use ``None`` for
            an infinite timeout.
        using (str, default="default"): The database to use.
        **timedelta_kwargs: Keyword arguments to directly supply to
            datetime.timedelta to create an interval. E.g.
            ``pglock.timeout(seconds=1, milliseconds=100)``
            will create a timeout of 1100 milliseconds.

    Raises:
        django.db.utils.OperationalError: When a timeout occurs
        TypeError: When the timeout interval is an incorrect type
    """
    if timedelta_kwargs:
        timeout = dt.timedelta(**timedelta_kwargs)
    elif timeout is _unset:
        raise ValueError("Must supply a value to pglock.timeout")

    if timeout is not None:
        timeout = _cast_timeout(timeout)

        if not timeout:
            raise ValueError(
                "Must supply value greater than a millisecond to pglock.timeout or use ``None`` to"
                " reset the timeout."
            )
    else:
        timeout = dt.timedelta()

    if not hasattr(_timeout, "value"):
        _timeout.value = None

    old_timeout = _timeout.value
    _timeout.value = int(timeout.total_seconds() * 1000)

    try:
        with connections[using].cursor() as cursor:
            cursor.execute(f"SELECT set_config('lock_timeout', '{_timeout.value}', false)")
            yield
    finally:
        _timeout.value = old_timeout

        with connections[using].cursor() as cursor:
            if not _is_transaction_errored(cursor):
                if _timeout.value is None:
                    cursor.execute("SELECT set_config('lock_timeout', NULL, false)")
                else:
                    cursor.execute(f"SELECT set_config('lock_timeout', '{_timeout.value}', false)")


def _cast_lock_id(lock_id):
    """Cast a lock ID into the appropriate type"""
    if isinstance(lock_id, str):
        return int.from_bytes(
            hashlib.sha256(lock_id.encode("utf-8")).digest()[:8], "little", signed=True
        )
    elif isinstance(lock_id, int):
        return lock_id
    else:
        raise TypeError(f'Lock ID "{lock_id}" is not a string or int')


def advisory_id(lock_id):
    """
    Given a lock ID, return the (classid, objid) tuple that Postgres uses
    for the advisory lock in the pg_locks table
    """
    lock_id = _cast_lock_id(lock_id)
    return lock_id >> 32, lock_id & 0xFFFFFFFF


class advisory(contextlib.ContextDecorator):
    """Obtain an advisory lock.

    Args:
        lock_id (Union[str, int], default=None): The ID of the lock. When
            using the decorator, it defaults to the full module path and
            function name of the wrapped function. It must be supplied to
            the context manager.
        shared (bool, default=False): When ``True``, creates a shared
            advisory lock. Consult the
            `Postgres docs <https://www.postgresql.org/docs/current/explicit-locking.html#ADVISORY-LOCKS>`__
            for more information.
        using (str, default="default"): The database to use.
        timeout (Union[int, float, datetime.timedelta, None]): Set a timeout when waiting
            for the lock. This timeout only applies to the lock acquisition statement and not the
            wrapped code. If 0, ``pg_try_advisory_lock`` will be used to return immediately. If
            ``None``, an infinite timeout will be used. When
            using a timeout, the acquisition status will be returned when running as a context
            manager. Use the ``side_effect`` argument to change the runtime behavior.
        side_effect (str): Adjust the runtime behavior when using a timeout.
            `pglock.Return` will return the acquisition status when using the context manager.
            `pglock.Raise` will raise a ``django.db.utils.OperationalError`` if the lock cannot
            be acquired or a timeout happens. `pglock.Skip` will skip decoratored code if the
            lock cannot be acquired. Defaults to `pglock.Return` when used as a context manager
            or `pglock.Raise` when used as a decorator.

    Returns:
        bool: When using the default side effect, returns ``True`` if the lock was acquired or
        ``False`` if not.

    Raises:
        django.db.utils.OperationalError: When a lock cannot be acquired or a timeout happens
            when using ``side_effect=pglock.Raise``.
        ValueError: If an invalid ``side_effect`` is provided or no lock ID is supplied for
            the context manager.
        TypeError: If the lock ID is not a string or int.
    """  # noqa

    def __init__(
        self,
        lock_id=None,
        *,
        shared=False,
        using=DEFAULT_DB_ALIAS,
        timeout=_unset,
        side_effect=None,
    ):
        """Acquire an advisory lock"""
        self.lock_id = lock_id
        self.using = using
        self.side_effect = side_effect
        self.shared = shared
        self.timeout = _cast_timeout(timeout)

        # Use pg_try_advisory.. when a timeout of 0 has been applied.
        self.nowait = isinstance(self.timeout, dt.timedelta) and not self.timeout

        # "_func" will be set if a function is wrapped
        self._func = None

    @property
    def int_lock_id(self):
        return _cast_lock_id(self.lock_id)

    @property
    def acquire(self):
        return f'pg{"_try" if self.nowait else ""}_advisory_lock{"_shared" if self.shared else ""}'

    @property
    def release(self):
        return f'pg_advisory_unlock{"_shared" if self.shared else ""}'

    def __call__(self, func):
        self._func = func

        @functools.wraps(func)
        def inner(*args, **kwargs):
            with self._recreate_cm():
                if self.acquired or self.side_effect != Skip:
                    return func(*args, **kwargs)

        return inner

    def _process_runtime_parameters(self):
        """
        Instantiate parameters such as the lock ID and side effect after
        we know if we are running as a context manager or decorator
        """
        if not self.lock_id and self._func:
            module = inspect.getmodule(self._func)
            self.lock_id = f"{module.__name__}.{self._func.__name__}"

        if self.side_effect is None:
            self.side_effect = Raise if self._func else Return
        else:
            self.side_effect = (
                self.side_effect.__class__
                if not inspect.isclass(self.side_effect)
                else self.side_effect
            )

        if not self.lock_id:
            raise ValueError("Must supply a lock ID as the first argument.")

        if self.side_effect not in (Return, Raise, Skip):
            raise ValueError(
                "side_effect must be one of pglock.Return, pglock.Raise, or pglock.Skip"
            )

        if not self._func and self.side_effect == Skip:
            raise ValueError(
                "Cannot use pglock.Skip in a context manager."
                " Use the @pglock.advisory decorator instead."
            )
        elif self._func and self.side_effect == Return:
            raise ValueError(
                "Cannot use pglock.Return in a @pglock.advisory() decorator."
                " Use it as a context manager instead."
            )

    def __enter__(self):
        self._process_runtime_parameters()

        sql = f"SELECT {self.acquire}({self.int_lock_id})"

        with connections[self.using].cursor() as cursor:
            try:
                with contextlib.ExitStack() as stack:
                    if self.timeout is not _unset and not self.nowait:
                        stack.enter_context(lock_timeout(self.timeout, using=self.using))

                        if self.side_effect != Raise and connections[self.using].in_atomic_block:
                            # If returning True/False, create a savepoint so that
                            # the transaction isn't in an errored state when returning.
                            stack.enter_context(transaction.atomic(using=self.using))

                    cursor.execute(sql)
                    self.acquired = cursor.fetchone()[0] if self.nowait else True
            except OperationalError:
                # This block only happens when the lock times out
                if self.side_effect != Raise:
                    self.acquired = False
                else:
                    raise

            if not self.acquired and self.side_effect == Raise:
                raise OperationalError(f'Could not acquire lock "{self.lock_id}"')

            self.stack = contextlib.ExitStack()
            if self.acquired and connections[self.using].in_atomic_block:
                # Create a savepoint so that we can successfully release
                # the lock if the transaction errors
                self.stack.enter_context(transaction.atomic(using=self.using))

            self.stack.__enter__()

            return self.acquired

    def __exit__(self, exc_type, exc_value, traceback):
        self.stack.__exit__(exc_type, exc_value, traceback)

        if self.acquired:
            with connections[self.using].cursor() as cursor:
                cursor.execute(f"SELECT {self.release}({self.int_lock_id})")


def model(
    *models,
    mode=ACCESS_EXCLUSIVE,
    using=DEFAULT_DB_ALIAS,
    timeout=_unset,
    side_effect=Return,
):
    """Lock model(s).

    Args:
        *models (Union[str, models.Model]): Model paths (e.g. "app_label.Model") or
            classes to lock.
        mode (str, default=pglock.ACCESS_EXCLUSIVE): The lock mode.
            See the
            `Postgres docs <https://www.postgresql.org/docs/current/explicit-locking.html#LOCKING-TABLES>`__
            for a list of all modes and what they mean.
            There is a constant for each one in the ``pglock`` module, e.g.
            ``pglock.ACCESS_SHARE``.
        using (str, default="default"): The database to use.
        timeout (Union[int, float, datetime.timedelta, None]): Set a timeout when waiting
            for the lock. If 0, ``NOWAIT` will be used to return immediately. If ``None``, the
            timeout is infinite. When using a timeout, the acquisition status will be returned.
            Use the ``side_effect`` argument to change the runtime behavior.
        side_effect (str, default=pglock.Return): Adjust the runtime behavior when using a timeout.
            `pglock.Return` will return the acquisition status. `pglock.Raise` will
            raise a ``django.db.utils.OperationalError`` if the lock cannot
            be acquired or a timeout happens.

    Returns:
        bool: When using the default side effect, returns ``True`` if the lock was acquired or
        ``False`` if not.

    Raises:
        django.db.utils.OperationalError: If ``side_effect=pglock.Raise`` and a lock cannot be
            acquired or a timeout occurs.
        RuntimeError: When running code outside of a transaction.
        ValueError: When ``side_effect`` is an invalid value or no models are supplied.
        TypeError: When ``timeout`` is an invalid type.

    """  # noqa
    timeout = _cast_timeout(timeout)
    side_effect = side_effect.__class__ if not inspect.isclass(side_effect) else side_effect

    if side_effect not in (Return, Raise):
        raise ValueError("side_effect must be one of pglock.Return or pglock.Raise")

    if not models:
        raise ValueError("Must supply at least one model to pglock.model().")

    if not connections[using].in_atomic_block:
        raise RuntimeError(f'Database "{using}" must be in a transaction to lock models.')

    # Use NOWAIT when a timeout of 0 has been applied.
    nowait = isinstance(timeout, dt.timedelta) and not timeout
    models = [apps.get_model(model) if isinstance(model, str) else model for model in models]
    models = ", ".join(f'"{model._meta.db_table}"' for model in models)
    sql = f'LOCK TABLE {models} IN {mode} MODE {"NOWAIT" if nowait else ""}'

    try:
        with contextlib.ExitStack() as stack:
            if side_effect == Return:
                # If returning True/False, create a savepoint so that
                # the transaction isn't in an errored state when returning.
                stack.enter_context(transaction.atomic(using=using))

            # Set the lock timeout when either ``None`` or a non-zero
            # timeout has been supplied.
            if timeout is not _unset and not nowait:
                stack.enter_context(lock_timeout(timeout, using=using))

            with connections[using].cursor() as cursor:
                cursor.execute(sql)
                return True
    except OperationalError:
        if side_effect == Return:
            return False
        elif side_effect == Raise:
            raise
        else:
            raise AssertionError


def _prioritize_bg_task(backend_pid, side_effect):
    if not side_effect:  # pragma: no cover
        return

    BlockedPGLock = apps.get_model("pglock.BlockedPGLock")
    qset = BlockedPGLock.objects.pid(backend_pid)
    side_effect(qset)


class _PeriodicTimer(threading.Timer):
    def run(self):
        self.exc = None
        try:
            while not self.finished.wait(self.interval):
                self.function(*self.args, **self.kwargs)
        except BaseException as e:
            self.exc = e

    def cancel(self):
        super().cancel()

        if self.exc:
            raise RuntimeError("Exception raised in side effect") from self.exc


class PrioritizeSideEffect(SideEffect):
    """Base class for `pglock.prioritize` side effects.

    Must override the ``worker`` method, which takes
    a `pglock.models.BlockedPGLock` queryset of all locks
    that are blocking the prioritized process.

    Return the process IDs or blocked locks that were
    handled.

    Prioritize side effects take optional filters
    when initialize, which are passed to the underlying
    `pglock.models.BlockedPGLock` queryset.
    """

    def __init__(self, **filters):
        self.filters = filters

    def worker(self, blocked_locks):
        raise NotImplementedError

    def __call__(self, blocked_locks):
        return self.worker(blocked_locks.filter(**self.filters))


class Terminate(PrioritizeSideEffect):
    """
    The side effect for terminating blocking locks
    when using `pglock.prioritize`.

    Calls ``teminate_blocking_activity`` on the
    blocked lock queryset.

    Supply a duration to only terminate queries lasting greater than the
    duration.
    """

    def worker(self, blocked_locks):
        return blocked_locks.terminate_blocking_activity()


class Cancel(PrioritizeSideEffect):  # pragma: no cover
    """
    The side effect for canceling blocking locks
    when using `pglock.prioritize`.

    Calls ``cancel_blocking_activity`` on the
    blocked lock queryset.

    Supply a duration to only cancel queries lasting greater than the
    duration.
    """

    def worker(self, blocked_locks):
        return blocked_locks.cancel_blocking_activity()


@contextlib.contextmanager
def prioritize(
    *,
    interval=1,
    periodic=True,
    using=DEFAULT_DB_ALIAS,
    retries=0,
    timeout=_unset,
    side_effect=Terminate,
):
    """Kill any blocking locks.

    `pglock.prioritize` has a periodic background worker thread that checks for blocking activity
    and terminates it.

    Args:
        interval (Union[int, float, datetime.timedelta], default=1): The interval at
            which the background worker runs. Defaults to running every second.
        periodic (bool, default=True): If the worker should be ran periodically. If False,
            blocking locks are only killed once after the initial interval has happened.
        using (str, default="default"): The database to use.
        timeout (Union[datetime.timdelta, int, float, None]): The lock timeout to apply to the
            wrapped code. This is synonymous with using with
            ``with pglock.prioritize(), pglock.timeout()``. Although the background worker should
            properly terminate blocking locks, this serves as a backup option
            to ensure wrapped code doesn't block for too long. Never use a ``timeout`` that
            is less than ``interval``.
        side_effect (pglock.PrioritizeSideEffect, default=pglock.Terminate): The side effect
            called by the background worker. Supplied a `BlockedPGLock` queryset of locks
            blocking the prioritized code. Returns a list of all blocking PIDs that have been
            handled. The default side effect of ``pglock.Terminte`` will terminate blocking
            processes. `pglock.Cancel` is another side effect that can be used to cancel
            blocking processes.

    Raises:
        django.db.utils.OperationalError: If ``timeout`` is used and the timeout expires.
    """
    side_effect = side_effect() if inspect.isclass(side_effect) else side_effect

    if isinstance(interval, (int, float)):
        interval = dt.timedelta(seconds=interval)

    if not isinstance(interval, dt.timedelta):
        raise TypeError('"interval" argument must be an int, float, or timedelta instance')

    backend_pid = pgactivity.pid(using=using)
    timer_class = _PeriodicTimer if periodic else threading.Timer
    killer_thread = timer_class(
        interval.total_seconds(), _prioritize_bg_task, (backend_pid, side_effect)
    )
    killer_thread.start()

    try:
        with contextlib.ExitStack() as stack:
            if timeout is not _unset:
                stack.enter_context(lock_timeout(timeout, using=using))

            yield
    finally:
        killer_thread.cancel()



"""Core way to access configuration"""
from django.conf import settings


def attributes():
    return getattr(
        settings,
        "PGLOCK_ATTRIBUTES",
        [
            "activity_id",
            "activity__duration",
            "granted",
            "mode",
            "rel_kind",
            "rel_name",
            "activity__context",
            "activity__query",
        ],
    )


def blocking_attributes():
    return getattr(
        settings,
        "PGLOCK_BLOCKING_ATTRIBUTES",
        [
            "activity_id",
            "blocking_activity_id",
            "activity__context",
            "blocking_activity__context",
            "activity__query",
            "blocking_activity__query",
        ],
    )


def configs():
    """Return pre-configured LS arguments"""
    return getattr(settings, "PGLOCK_CONFIGS", {})


def limit():
    """The default limit when using the pglock command"""
    return getattr(settings, "PGLOCK_LIMIT", 25)


def get(name, **overrides):
    """Get a configuration with overrides"""
    if not name:
        cfg = {}
    elif name not in configs():
        raise ValueError(f'"{name}" is not a valid config name from settings.PGLOCK_CONFIGS')
    else:
        cfg = configs()[name]

    # Note: We might allow overriding with "None" or empty values later, but currently no
    # settings allow this. This code filters overrides so that management commands can
    # simply pass in options that might already contain Nones
    cfg.update(**{key: val for key, val in overrides.items() if val})

    if "limit" not in cfg:
        cfg["limit"] = limit()

    if "attributes" not in cfg:
        cfg["attributes"] = attributes() if not cfg.get("blocking") else blocking_attributes()

    return cfg



import datetime as dt
import os
import re
import sys
import textwrap

from django.core.management.base import BaseCommand
from django.db.models import F

from pglock import config, models


def get_terminal_width():  # pragma: no cover
    try:
        return os.get_terminal_size().columns
    except OSError:  # This only happens during testing
        return 80


def _format(val, expanded):
    if isinstance(val, dt.timedelta):
        if val:  # pragma: no branch
            val -= dt.timedelta(microseconds=val.microseconds)
    elif isinstance(val, str):
        if not expanded:
            val = " ".join(val.split())
        else:
            val = textwrap.dedent(val).strip()

    return str(val)


def _handle_user_input(*, cfg, num_locks, stdout):
    is_blocking = cfg.get("blocking")
    is_cancel = cfg.get("cancel")

    if not num_locks:
        stdout.write(
            f"No {'blocking ' if is_blocking else ''}queries"
            f" to {'cancel' if is_cancel else 'terminate'}."
        )
        return False

    if not cfg.get("yes"):
        pluralize = "y" if num_locks == 1 else "ies"
        resp = input(
            (
                f"{'Cancel' if is_cancel else 'Terminate'} "
                f"{num_locks} {'blocking ' if is_blocking else ''}quer{pluralize}? (y/[n]) "
            )
        )
        if not re.match("^(y)(es)?$", resp, re.IGNORECASE):
            stdout.write("Aborting!")
            return False

    return True


class Command(BaseCommand):
    help = "Show and manage locks."

    def add_arguments(self, parser):
        parser.add_argument("pids", nargs="*", type=str)
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-f",
            "--filter",
            action="append",
            dest="filters",
            help="Filters for the underlying queryset",
        )
        parser.add_argument(
            "-o",
            "--on",
            action="append",
            dest="on",
            help="Show locks on models",
        )
        parser.add_argument(
            "-a",
            "--attribute",
            action="append",
            dest="attributes",
            help="Attributes to show",
        )
        parser.add_argument("--blocking", action="store_true", help="Show blocking locks")
        parser.add_argument("-l", "--limit", help="Limit results")
        parser.add_argument("-e", "--expanded", action="store_true", help="Show an expanded view")
        parser.add_argument("-c", "--config", help="Use a config from settings.PGLOCK_CONFIGS")
        parser.add_argument("-y", "--yes", action="store_true", help="Don't prompt for input")

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--cancel",
            action="store_true",
            help="Cancel activity",
        )
        group.add_argument(
            "--terminate",
            action="store_true",
            help="Terminate activity",
        )

    def handle(self, *args, **options):
        cfg = config.get(options["config"], **options)
        is_blocking = cfg.get("blocking")
        is_cancel = cfg.get("cancel")
        is_terminate = cfg.get("terminate")

        if cfg.get("blocking"):
            qset = models.BlockedPGLock.objects.select_related("activity", "blocking_activity")
        else:
            qset = models.PGLock.objects.select_related("activity")

        locks = qset.config(options["config"], **options).values(*cfg["attributes"])

        term_w = get_terminal_width()
        expanded = cfg.get("expanded", False)

        if is_cancel or is_terminate:
            locks = (
                locks.distinct("activity_id")
                if not is_blocking
                else locks.distinct("blocking_activity_id")
            )
            num_locks = len(locks)

            if not _handle_user_input(cfg=cfg, num_locks=num_locks, stdout=self.stdout):
                sys.exit(1)

            method_name = (
                f"{'cancel' if is_cancel else 'terminate'}"
                f"{'_blocking' if is_blocking else ''}_activity"
            )
            num_success = len(getattr(locks, method_name)())
            pluralize = "y" if num_success == 1 else "ies"
            self.stdout.write(
                (
                    f"{'Canceled' if is_cancel else 'Terminated'} "
                    f"{num_success} {'blocking ' if is_blocking else ''}quer{pluralize}"
                )
            )
        else:
            locks = locks.order_by(
                "granted",
                # Only PG14 and up has a non-null wait_duration. Sort by this first, but
                # use activity__duration as a backup if wait_duration is null
                F("wait_duration").desc(nulls_last=True),
                F("activity__duration").desc(nulls_last=True),
            )

            if not cfg.get("pids") and cfg.get("limit"):
                locks = locks[: cfg["limit"]]

            for lock in locks:
                if cfg.get("expanded"):
                    self.stdout.write("\033[1m" + "─" * term_w + "\033[0m")
                    for a in cfg["attributes"]:
                        self.stdout.write(f"\033[1m{a}\033[0m: {_format(lock[a], expanded)}")
                else:
                    line = " | ".join(_format(lock[a], expanded) for a in cfg["attributes"])
                    line = line[:term_w]
                    self.stdout.write(line)



from datetime import timedelta

from pglock.core import (
    ACCESS_EXCLUSIVE,
    ACCESS_SHARE,
    advisory,
    advisory_id,
    Cancel,
    EXCLUSIVE,
    lock_timeout as timeout,
    model,
    prioritize,
    PrioritizeSideEffect,
    Raise,
    Return,
    ROW_EXCLUSIVE,
    ROW_SHARE,
    SHARE,
    SHARE_ROW_EXCLUSIVE,
    SHARE_UPDATE_EXCLUSIVE,
    SideEffect,
    Skip,
    Terminate,
)
from pglock.version import __version__


__all__ = [
    "ACCESS_EXCLUSIVE",
    "ACCESS_SHARE",
    "advisory",
    "advisory_id",
    "prioritize",
    "Cancel",
    "EXCLUSIVE",
    "model",
    "prioritize",
    "PrioritizeSideEffect",
    "Raise",
    "Return",
    "ROW_EXCLUSIVE",
    "ROW_SHARE",
    "SHARE",
    "SHARE_ROW_EXCLUSIVE",
    "SHARE_UPDATE_EXCLUSIVE",
    "SideEffect",
    "Skip",
    "Terminate",
    "timedelta",
    "timeout",
    "__version__",
]



from django.core.exceptions import ImproperlyConfigured
from django.utils.version import get_version_tuple


def _psycopg_version():
    try:
        import psycopg as Database
    except ImportError:
        import psycopg2 as Database
    except ImportError:  # pragma: no cover
        raise ImproperlyConfigured("Error loading psycopg2 or psycopg module")

    version_tuple = get_version_tuple(Database.__version__.split(" ", 1)[0])

    if version_tuple[0] not in (2, 3):  # pragma: no cover
        raise ImproperlyConfigured(f"Pysocpg version {version_tuple[0]} not supported")

    return version_tuple


psycopg_version = _psycopg_version()
psycopg_maj_version = psycopg_version[0]


def pg_maj_version(cursor):
    """Return the major version of Postgres that's running"""
    version = getattr(cursor.connection, "server_version", cursor.connection.info.server_version)
    return int(str(version)[:-4])
